#! /usr/bin/env python
# -*- coding: utf-8 -*-
from . import hera
from . import json
from .sql import mysql
from lxml import etree

MODULES = {
    'xml:hera': hera,
    'json:hera': json,
    'sql:mysql': mysql,
    }


def getDatabase(db):
    get = getattr(MODULES[db.format], 'getDatabase')
    tree = get(db)
    return tree


def getItems(tree, filter=None):
    return _get_nodes(tree, 'item', filter)


def getItem(tree, filter=None):
    return _get_node(tree, 'item', filter)


def getMetadata(item, pid=None):
    if pid:
        xpath = './/metadata[@pid="{pid}"]'.format(pid=pid)
        return item.findall(xpath)
    return item.findall('.//metadata')


def getValues(item, pid):
    xpath = './/metadata[@pid="{pid}"]'.format(pid=pid)
    values = []
    for metadata in item.findall(xpath):
        values.append(metadata.text)
    return values


def getValue(item, pid):
    values = getValues(item, pid)
    if len(values) == 0:
        return None
    if len(values) == 1:
        return values[0]
    raise IndexError("Too many metadata ({count})".format(count=len(values)))


def getProperties(tree, filter=None):
    return _get_nodes(tree, 'property', filter)


def getProperty(tree, filter=None):
    return _get_node(tree, 'property', filter)


def createProperty(tree, pid, **kwargs):
    properties = tree.findall('.//properties')
    if properties:
        properties = properties[0]
        props = properties.getchildren()
    else:  # Create <properties/> node
        root = tree.find('//hera')
        properties = etree.SubElement(root, 'properties')
        props = []
    # Check property unique id {pid} does not exist
    if len([p for p in props if p.get('id') == pid]) > 0:
        raise ValueError(f"Property id '{pid}' already exists")
    # Create property
    p = etree.SubElement(properties, 'property', id=pid)
    type = kwargs.get('type', 'text')
    etree.SubElement(p, 'type').text = type
    param = kwargs.get('name', None)
    if param:
        etree.SubElement(p, 'name').text = param
    param = kwargs.get('description', None)
    if param:
        etree.SubElement(p, 'description').text = param
    param = kwargs.get('uri', [])
    for value in param:
        etree.SubElement(p, 'uri').text = value
    return p


def deleteProperty(tree, filter=None):
    node = _get_node(tree, 'property', filter)
    node.getparent().remove(node)


def delete_unused_properties(tree):
    # give ourselves a map of unused properties, initialized with all of them
    properties = {}
    for p in getProperties(tree):
        properties[p.get('id')] = p
    # let's check which properties are really unused
    for e in getEntities(tree):
        for a in getAttributes(e):
            pid = a.get('pid')
            if pid in properties.keys():
                # this property is in use ; so we mustn't delete it
                properties.pop(pid)
    # NOTE we could check items metadata for usage, too
    #      however, we won't for now, for two reasons:
    #      1. because of performance considerations
    #      2. because HERA allows orphan metadata, and that's exactly
    #         what would be metadata referencing a pid which is not
    #         used by any of the attributes of this item's entity.
    #      This excludes items from non-relational databases, tho ...

    # the map now only contains unused ones ; so, delete them
    for pid, p in properties.items():
        p.getparent().remove(p)
    # end of function, don't bother with properties.clear()


def merge_properties(tree, props):
    for item in getItems(tree):
        for old, now in props.items():
            metadata = getMetadata(item, pid=old)
            for m in metadata:
                m.set('pid', now)


def getEntities(tree, filter=None):
    return _get_nodes(tree, 'entity', filter)


def getEntity(tree, filter=None):
    return _get_node(tree, 'entity', filter)


def deleteEntity(tree, filter=None):
    node = _get_node(tree, 'entity', filter)
    node.getparent().remove(node)


def getAttributes(entity, filter=None):
    return _get_nodes(entity, 'attribute', filter)


def getAttribute(entity, filter=None):
    return _get_node(entity, 'attribute', filter)


def createAttribute(entity, **kwargs):
    pid = kwargs['pid']
    a = etree.SubElement(entity, 'attribute', pid=pid)

    def _maybe_set_value(key, default=None):
        param = kwargs.get(key, default)
        if param is not None:
            a.set(key, str(param))

    _maybe_set_value('id')
    _maybe_set_value('min', 0)
    _maybe_set_value('max')

    def _maybe_add_child(key, default=None):
        param = kwargs.get(key, default)
        if param:
            etree.SubElement(a, key).text = param

    _maybe_add_child('type')
    _maybe_add_child('name')
    _maybe_add_child('description')
    _maybe_add_child('uri')

    return a


def updateAttribute(attribute, **kwargs):

    def _maybe_set_value(key):
        try:
            value = kwargs[key]
            if value is not None:
                attribute.set(key, str(value))
            else:  # remove value
                attribute.attrib.pop(key)
        except KeyError:
            pass  # nothing to change

    _maybe_set_value('id')
    _maybe_set_value('pid')
    _maybe_set_value('min')
    _maybe_set_value('max')

    def _maybe_update_child(key):
        try:
            value = kwargs[key]
            children = _get_nodes(attribute, key)
            for child in children:
                if value is not None:
                    child.text = str(value)
                else:  # remove child node
                    child.getparent().remove(child)
        except KeyError:
            pass  # nothing to change

    _maybe_update_child('type')
    _maybe_update_child('name')
    _maybe_update_child('description')
    _maybe_update_child('uri')


def refactor_relationship(tree, relationship, eid, euid, pid, cleanup=True):
    # TODO only create property if not exists
    p = createProperty(tree, pid=pid)
    e = getEntity(tree, lambda n: n.get('id') == eid)
    # TODO only create attribute if not exists
    a = createAttribute(e, pid=pid)
    # iterate over all items belonging to the relationship entity
    for old in getItems(tree, lambda n: n.get('eid') == relationship.eid):
        source = getMetadata(old, relationship.source)[0].text
        target = getMetadata(old, relationship.target)[0].text

        def is_source(item):
            is_of_entity = item.get('eid') == eid
            has_unique_id = False
            # Unique id shouldn't be a repeatable attribute,
            # but we know what real world looks like. Thus,
            # try to not break in this edge case, and let's
            # hope our caller knows what she does.
            for v in getValues(item, euid):
                has_unique_id = has_unique_id or (v == source)
            return is_of_entity and has_unique_id

        # get the item which must contain the new repeatable metadata
        now = getItem(tree, is_source)
        etree.SubElement(now, 'metadata', pid=pid).text = target
        if cleanup:
            # delete `old` relationship item, because it is
            # now represented by new metadata in item `now`
            old.getparent().remove(old)
    if cleanup:
        # delete the `relationship` entity, as there are no more items using it
        deleteEntity(tree, lambda n: n.get('id') == relationship.eid)


def merge_l10n_attributes(tree, entity, attr_vs_language, base_aid, cleanup=True):
    base = getAttribute(entity, lambda a: a.get('id') == base_aid)
    amin = 0
    amax = 1
    for aid, language in attr_vs_language.items():
        attribute = getAttribute(entity, lambda a: a.get('id') == aid)
        atype = _get_node(attribute, 'type')
        atype = atype.text if atype is not None else 'text'
        if atype != 'text':
            raise ValueError(f"Attribute '{aid}':'{type}' not localized")
        amin = max(amin, int(attribute.get('min')))
        amax = max(amax, int(attribute.get('max')))
        node = _get_node(attribute, 'name')
        qname = etree.QName('http://www.w3.org/XML/1998/namespace', 'lang')
        if node is not None:
            node.set(qname, language)
            base.append(node)
        node = _get_node(attribute, 'description')
        if node is not None:
            node.set(qname, language)
            base.append(node)

        pid = attribute.get('pid')
        for item in getItems(tree):
            for metadata in getMetadata(item):
                if metadata.get('pid') == pid:
                    metadata.set('pid', base.get('pid'))
                    metadata.set(qname, language)

        if not cleanup or aid == base_aid:
            continue
        attribute.getparent().remove(attribute)


def _get_nodes(tree, tag, filter=None):
    nodes = tree.findall(f'.//{tag}')
    if filter:
        return [node for node in nodes if filter(node)]
    return nodes


def _get_node(tree, tag, filter=None):
    nodes = _get_nodes(tree, tag, filter)
    if len(nodes) == 0:
        return None
    if len(nodes) == 1:
        return nodes[0]
    raise IndexError(f"Too many {tag} elements ({len(nodes)})")


class Relationship:
    def __init__(self, eid, source, target):
        self.eid = eid  # a Relationship is an Entity, it's its id
        self.source = source  # source attribute pid
        self.target = target  # target attribute pid


class Config:
    def __init__(self, **entries):
        self.__dict__.update(entries)
