# -*- coding: utf-8 -*-
import pytest
import heimdall

@pytest.fixture
def setup_data():
    args = {
        'format': 'sql:mysql',
        'url': 'mysql://root:password@localhost:3306/menagerie',
        'entities': ('pet', ),
        }
    config = heimdall.Config(**args)

    #print("\nSetting up…")
    yield config
    # Teardown: Clean up resources (if any) after the test
    #print("\nTearing down…")


@pytest.mark.parametrize(
    ('attr', 'expected'),
    [
        ('scheme', 'mysql'),
        ('username', 'root'),
        ('password', 'password'),
        ('hostname', 'localhost'),
        ('port', 3306),
        ('path', '/menagerie'),
    ],
)
def test_smoke_url_format(setup_data, attr, expected):
    from urllib.parse import urlparse
    url = urlparse(setup_data.url)
    assert getattr(url, attr) == expected


@pytest.mark.parametrize(
    ('table','row', 'type','min','max'),
    [
     (
        'test_table',
        ('name', 'varchar(20)', 'utf8mb4_0900_ai_ci', 'NO', '', None, '',
         'select,insert,update,references', 'Name of this adorable animal'),
        'text', 1, 1
     ),
     (
        'test.table',
        ('sex', 'char(1)', 'utf8mb4_0900_ai_ci', 'YES', '', None, '',
         'select,insert,update,references', ''),
        'text', 0, 1
     ),
     (
         'blabla',
         ('birth', 'date', None, 'YES', '', None, '',
          'select,insert,update,references', '出生日期'),
        'datetime', 0, 1
     ),
    ],
)
def test_create_attribute(setup_data, table,row, type,min,max):
    from heimdall.sql.mysql import _create_attribute
    result = _create_attribute(table, row)
    assert result.tag == 'attribute'
    assert len(result.attrib) == 4
    assert result.get('id') == f'{table}.{row[0]}_attr'
    assert result.get('pid') == f'{table}.{row[0]}'
    assert result.get('min') == str(min)
    assert result.get('max') == str(max)
    assert get_child(result, 'type').text == type
    assert get_child(result, 'name').text == row[0]
    node = get_child(result, 'description')
    assert node is None or node.text == row[8]

@pytest.mark.parametrize(
    ('table','row', 'type'),
    [
     (
        'test_table',
        ('name', 'varchar(20)',
         None, None, None, None, None, None,  # these are unused
         'Name of this adorable animal'),
        'text',
     ),
     (
        'test.table',
        ('sex', 'char(1)',
         None, None, None, None, None, None,  # these are unused
         ''),
        'text',
     ),
     (
         'fzqs',
         ('birth', 'date',
         None, None, None, None, None, None,  # these are unused
          '出生日期'),
        'datetime',
     ),
    ],
)
def test_create_property(setup_data, table,row, type):
    from heimdall.sql.mysql import _create_property
    result = _create_property(table, row)
    assert result.tag == 'property'
    assert len(result.attrib) == 1
    assert result.get('id') == f'{table}.{row[0]}'
    assert get_child(result, 'type').text == type
    assert get_child(result, 'name').text == row[0]
    node = get_child(result, 'description')
    assert node is None or node.text == row[8]

def test_create_entity(setup_data):
    from heimdall.sql.mysql import _create_entity
    from urllib.parse import urlparse
    url = urlparse(setup_data.url)
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
            user=url.username, password=url.password,
            host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        for table in setup_data.entities:

            result, properties = _create_entity(table, cursor)
            assert result.tag == 'entity'
            assert len(result.attrib) == 1
            assert result.get('id') == table
            assert len(result.getchildren()) == 9
            assert len(result.getchildren()) == len(properties)+2
            tags = ['name', 'description', 'attribute', ]
            for child in result.iterchildren():
                assert child.tag in tags
                if child.tag == 'name':
                    assert child.text == table
                elif child.tag == 'attribute':
                    assert len(child.attrib) == 4
                    assert child.get('id').startswith(table)
                    assert child.get('id').endswith('_attr')
                    assert child.get('pid').startswith(table)
                    name = child.get('pid').split('.')[-1]
                    assert child.get('min') is not None
                    assert child.get('max') is not None
                    assert get_child(child, 'type').text is not None
                    assert get_child(child, 'name').text == name
                    d = get_child(child, 'description')
                    assert d is None or d.text is not None
    connection.close()

def test_create_tree(setup_data):
    from heimdall.sql.mysql import _create_tree
    from urllib.parse import urlparse
    url = urlparse(setup_data.url)
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
            user=url.username, password=url.password,
            host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        tree = _create_tree(setup_data.entities, cursor)
        assert len(tree.getchildren()) == 3  # items, entities, properties
        items = tree.xpath('//item')
        assert len(items) == 9
        for item in items:
            metadata = item.getchildren()
            assert len(metadata) <= 7
            for meta in metadata:
                assert meta.tag == 'metadata'
        entities = tree.xpath('//entity')
        assert len(entities) == 1
        properties = tree.xpath('//property')
        assert len(properties) == 7

    connection.close()


def test_get_database(setup_data):
    from urllib.parse import urlparse
    url = urlparse(setup_data.url)
    #from .logger import getLogger
    #log = getLogger(__name__)
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
            user=url.username, password=url.password,
            host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        for table in setup_data.entities:

            result = cursor.execute(f'SHOW FULL COLUMNS FROM {table}')
            rows = cursor.fetchall()
            #for row in rows:
            #   log.info(row)
            #log.info('-----')

            result = cursor.execute(f'SELECT * FROM {table}')
            rows = cursor.fetchall()
            #for row in rows:
            #   log.info(row)
    connection.close()

def get_children(node, tag):
    return [n for n in node.getchildren() if n.tag == tag]

def get_child(node, tag):
    children = get_children(node, tag)
    if len(children) == 0:
        return None
    if len(children) == 1:
        return children[0]
    raise AssertionError(f"Expected max 1 '{tag}' child, got {len(children)}")
